#include <stdio.h>
#include <stdlib.h>
#include <sys/io.h>
#include <sys/types.h>
#include <fcntl.h>

#define DATA (0x378u) /* LPT1 = $378 */
#define STATUS (DATA+1u)
#define CONTROL (DATA+2u)
#define WAIT_USEC (50u)   /* hold signal for usecs */
#define PRINTER_SELECT_BIT (1u << 3u)
#define ACK_BIT (1u << 6u)

#define XPLODER_CMD_INIT (0x57u)
#define XPLODER_CMD_FREEZE (0x4Cu)
#define XPLODER_CMD_UNFREEZE (0x52u)

typedef unsigned char byte_t;
typedef FILE file_t;

void printer_select_high(void);
void printer_select_low(void);
void wait_for_ack_high(void);
void wait_for_ack_low(void);

int main(int argc, char *argv[]);

/* printer select to TRUE */
/* inverse logic, so FALSE actually */
void printer_select_high(void)
{
  volatile byte_t r = 0u;
  
  /* read current state from control port */
  r  = inb(CONTROL);
  /* set bit 3 to FALSE */
  r = (r & ~PRINTER_SELECT_BIT);
  /* set this state to the control port */
  outb(r, CONTROL);
  /* hold it for a few usecs */
  usleep(WAIT_USEC);
}

/* printer select to FALSE */
/* inverse logic, so TRUE actually */
void printer_select_low(void)
{ 
  volatile byte_t r = 0u;
  
  /* read current state from control port */
  r = inb(CONTROL);
  /* set bit 3 to TRUE */
  r = (r | PRINTER_SELECT_BIT);
  /* set this state to the control port */
  outb(r, CONTROL);
  /* hold it for a few usecs */
  usleep(WAIT_USEC);
}

/* wait for the ack bit to become TRUE */
void wait_for_ack_high(void)
{
  volatile byte_t r = 0u;

  do
  {
    /* read current state from status port */
    r = inb(STATUS);
  }
  /* while bit 6 of status port is FALSE */
  while(0u == (r & ACK_BIT));
}

/* wait for the ack bit to become FALSE */
void wait_for_ack_low(void)
{
  volatile byte_t r = 0u;

  do
  {
    /* read current state from status port */
    r = inb(STATUS);
  }
  /* while bit 6 of status port is TRUE */
  while(ACK_BIT == (r & ACK_BIT));
}

/* send a single byte */
void send_byte(byte_t byte)
{
  /* set the output state */
  outb(byte, DATA);
  usleep(WAIT_USEC);
    
  /* set printer select to TRUE */
  printer_select_high();
  /* wait till ack becomes TRUE */
  wait_for_ack_high();
  /* set printer select to FALSE */
  printer_select_low();
  /* wait till ack becomes FALSE */
  wait_for_ack_low();
}

void cmd_freeze(void)
{
  /* hey xploder/xplorer */
  send_byte(XPLODER_CMD_INIT);
  /* we want to freeze */
  send_byte(XPLODER_CMD_FREEZE);
}

void cmd_unfreeze(void)
{
  /* hey xploder/xplorer */
  send_byte(XPLODER_CMD_INIT);
  /* we want to unfreeze */
  send_byte(XPLODER_CMD_UNFREEZE);
}

/* main */
int main(int argc, char *argv[]) 
{
  int tem = 0;
  /* the current byte from the file */
  byte_t b = 0u;
  /* for calculating the checksum */
  unsigned int sum = 0u;

  /* for remembering the state as we entered */
  byte_t data_b = 0u;
  byte_t status_b = 0u;
  byte_t control_b = 0u;
  /* for reading the state after we did everything */
  byte_t data_a = 0u;
  byte_t status_a = 0u;
  byte_t control_a = 0u;

  /* set permissions to access port */
  /* but we still need a privileged user to exec the program */
  if(ioperm(DATA, 3, 1)) 
  {
    perror("ioperm"); 
    return EXIT_FAILURE;
  }

  tem = fcntl(0, F_GETFL, 0);
  /*(void) fcntl(0, F_SETFL, (tem | O_NDELAY));*/

  /* remember the state before we do anything */
  data_b = inb(DATA);
  status_b = inb(STATUS);
  control_b = inb(CONTROL);  
  
  {
    char c = 0;
    ssize_t n = 0; 

    do
    {
      /* if a key is pressed => break out of loop */
      n = read(0, &c, 1);
      if(0 < n) 
      {
        /* let's talk to the xploder/xplorer */
        switch(c)
        {
          case 'F':
          case 'f': cmd_freeze();
                    #ifdef DEBUG 
                      (void) printf("Freeze\n"); 
                    #endif
                    break;
          case 'U':
          case 'u': cmd_unfreeze();
                    #ifdef DEBUG 
                      (void) printf("Unfreeze\n"); 
                    #endif
                    break;
          default:
                    break;
        }
      }
    }
    while(1);
  }
  
  /* get the current state and compare */
  data_a = inb(DATA);
  status_a = inb(STATUS);
  control_a = inb(CONTROL);  

  #ifdef DEBUG
    (void) printf("b(%02X)?=a(%02X)\n", data_b, data_a);
    (void) printf("b(%02X)?=a(%02X)\n", status_b, status_a);
    (void) printf("b(%02X)?=a(%02X)\n", control_b, control_a);
  #endif
    
  (void) fcntl(0, F_SETFL, tem);

  /* flush stdout */
  (void) printf("\n");
    
  /* take away permissions to access port */
  if(ioperm(DATA, 3, 0)) 
  {
    perror("ioperm"); 
    return EXIT_FAILURE;
  }
    
  return EXIT_SUCCESS;
}
